package org.apache.derby.impl.services.cache;

import java.io.FileWriter;
import java.io.IOException;
import java.util.*;

import org.apache.derby.iapi.error.StandardException;
import org.apache.derby.iapi.services.cache.*;
import org.apache.derby.iapi.services.context.ContextManager;
import org.apache.derby.iapi.services.daemon.DaemonService;
import org.apache.derby.iapi.services.daemon.Serviceable;
import org.apache.derby.iapi.services.sanity.SanityManager;
import org.apache.derby.iapi.store.raw.PageKey;
import org.apache.derby.iapi.util.Matchable;
import org.apache.derby.iapi.util.Operator;
import org.apache.derby.impl.store.raw.data.StoredPage;

/**
 * A LRU cache manager. Has one list for the buffer and one 
 * hashmap for quick lookup in the list. 
 * 
 *  Some methods have extra verification methods for debugging 
 *  purposes. Derby embedded debugging catches a lot of bugs.
 *  But this is often bundled in a lot of nested exception
 *  handling, and shutdown of the database. Finding the
 *  root (exeception) for these cascading events is not very easy.
 *  The verification methods are designed to let the designer
 *  know what exactly went wrong.
 *  
 * Pages in aIn and aOut are put in holder items called 
 * LRUItems. Most important methods are find, release 
 * and findfreeItem. 
*  Not all interface methods are implemented. As far as we
*  can see many methods never gets called in a page cache 
*  manager implementation, like this.
 *  
 * @see Servicable
 * @see CacheManager
 * @see LRUItem
 * @see Cacheable
 * 
 */


 
public class LRU extends HashMap
    implements CacheManager, Serviceable
{
    private String name;
    final long MAX_ENTRIES;
    public final CacheStat stat = new CacheStat();
    private LinkedList lru;
    private CacheableFactory holderFactory;
    private boolean active;
    private int usedItems;

    /// Fields only used for debuggin and statistics.
    private int counter=0;
    private int lineNr=0;
    
    FileWriter f;
    FileWriter lruDump;
    FileWriter findCount;
    int nrFind=0;
    private boolean debug = true;
    private boolean log = true;
    
    /**
     * Constructor - Currently ignores initial size and goes
     *  right to maximum size.
     * 
	 *	@param holderFactory the cacheable object class
	 *	@param name the name of the cache
	 *	@param initialSize the initial number of cachable 
	 *		object this cache holds.
	 *	@param maximumSize the maximum size of the cache.  

     */
    
    public LRU(CacheableFactory holderFactory, String name, int initialSize, long maximumSize, boolean useByteCount)
    {
    	System.out.println("TestLRU");
    	System.out.println("initialsize "+initialSize);
    	System.out.println("maximumsize "+maximumSize);
        this.name = name;
        MAX_ENTRIES = maximumSize;
        stat.initialSize = initialSize;
        stat.maxSize = maximumSize;
        this.holderFactory = holderFactory;
        initialize(maximumSize);
        active = true;
        usedItems =0;
        try{
        	lruDump = new FileWriter("lru.txt");
        	lruDump = new FileWriter("lru.txt",true);
        }catch (IOException e) {
			e.printStackTrace();	
		}
    }

    /**
     * 
     * Initializes the lru list and creates LRUItems.
     * 
     * @param initialsize - size of cache
     */
    private void initialize(long initialsize)
    {
        lru = new LinkedList();
        for(int i = 0; i < initialsize; i++)
        {
            LRUItem item = new LRUItem();
            lru.add(item);
        }

    }
    
   ////// Interface methods from CacheManager ///////////////////////////////////////
    

    /**
     * All items that will be accessed from the 
     * cachemanager must first be created by this method.
     * The page is put in the front of aIn.
     * 
     * @see PageKey
     * 
     * @param key - Pagekey of page to be created
     * @param createParameter
     */


    public Cacheable create(Object key, Object createParameter)
        throws StandardException
    {
    	if(debug)System.out.println("create");
        
    	LRUItem item = (LRUItem)get(key);
    	if(item!= null)	throw StandardException.newException("XBCA0.S", name, key);
    	
        Cacheable entry = null;
        Cacheable oldentry=null;
        findFreeItem(lru,(int)MAX_ENTRIES,this);
        
        lru.remove(item);
        lru.addFirst(item);
        
        entry = item.getEntry();
        if(entry != null)
        {
            if(entry.isDirty())
                entry.clean(false);
            if(entry.getIdentity()!=null)
            	entry.clearIdentity();
            
        } 
        
        entry = holderFactory.newCacheable(this);
        
        oldentry = entry;
        entry = createIdentity(oldentry,key,createParameter);
        item.incKeep();
        
        item.setEntry(entry);
        
        put(key, item);
        if(SanityManager.DEBUG)
        	return verifyCreate(entry,key,item);
        else return entry;
    }
    /**
     * Verification method for create, used for debugging
     * 
     * @param entry - from create
     * @param key - from create
     * @param item - from cre<ate
     * @return - Cacheable to be returned from create (if no errors are found}.
     * @throws StandardException
     */
    public Cacheable verifyCreate(Cacheable entry, Object key, LRUItem item) throws StandardException{
    	
    	int state = 0x000000000000000;
    	
    	if(entry==null) state=+ 1;
    	if(item.getEntry()!=entry) state=+ 2;
    	if(!entry.getIdentity().equals(key)) state=+ 4;
    	if(item.keepCount!=1) state=+ 8;
    	if(entry.getIdentity() == null) state +=16;
    	if(((LRUItem)get(key)).getEntry()!=entry) state+=32;
    	if(state!=0){
    		System.out.println("Create Errorstate: "+Integer.toBinaryString(state));
    		throw StandardException.newException("Error in create");
    	}
    	
    	return entry;
    	
    }
    
    /**
     * Find an object in cache or fault in from disk. 
     * Method contains, together with findfreeItem, the 
     * LRU logic. A call to find result in one of two cases.
     *
     * Case1 - Page found in cache. Page is put at front 
     * of lru que.
     * Case2 - Page is not found in cache. Page is faulted 
     * in and put at the front of lru que.
     * 
     *@param key - the key to the object
	 *@return a cacheable object that is kept in the cache.
	 *@exception StandardException Cloudscape Standard 
	 *error policy
	*/

    public Cacheable find(Object key)
        throws StandardException
    {
    	if(debug)System.out.println("find");
    	if(debug)nrFind++;
    	Cacheable entry;
    	LRUItem item;
    	
    	synchronized (this) {
    		
//    		 Debug code, writing statistics
    		counter++;
			if(counter == 10) {
				try {
					f = new FileWriter("lruStat.txt");
					f.write("" + stat+" \n");
					f.close();
					findCount= new FileWriter("findCount.txt");
					findCount.write("#find "+nrFind);
					findCount.close();
				} catch(Exception e){
					System.out.println("Error med skriving til fil");
					e.printStackTrace();
				}
				counter =0;
			}
			
			
			// If Derby is in shutdown return null
			
	        if(!active)
	            return null;
	        
	        item = (LRUItem)get(key);
			
	        if((item!=null)&&(item.getEntry()!=null)){
	        	entry=item.getEntry();
			}
	        
	        if(item == null)
	        {
	        	/// CASE 2 - Item not found in cache
	        	stat.findMiss++;
	        	item = findFreeItem(lru,(int)MAX_ENTRIES,this);
	            entry=item.getEntry();
	            entry =setIdentity(entry,key);
	            item.setEntry(entry);
	            lru.remove(item);
	            lru.addFirst(item);
	            item.incKeep();
	            put(key, item);
	        }
	        else{
	        	// CASE 1 - Item found
	        	stat.findHit++;
	        	entry = item.getEntry();
	        	lru.remove(item);
	        	lru.addFirst(item);
	        	item.incKeep();
	        }
	        if(SanityManager.DEBUG) return verifyFind(entry,item,key);
	    	else return entry;
    	}
    	
    }
    
    /**
      * Verification method for find.
     * 
     * @param entry - Cacheable to be returned 
     * @param item - LRUItem holding entry
     * @param key - Pagekey of entry
     * @return
     * @throws StandardException
     */
    public Cacheable verifyFind(Cacheable entry,LRUItem item,Object key) throws StandardException{
    
    	System.out.println("used items: "+usedItems);
    	try{
    		lruDump.write("\n");
    		Cacheable testEntry;
    		lruDump.write("line:"+lineNr+" ");
    		lineNr++;
    	synchronized (this) {
		
    		for (Iterator iter = lru.iterator(); iter.hasNext();) {
			LRUItem lruItem = (LRUItem) iter.next();
			testEntry=lruItem.getEntry();
			
			PageKey pk=null;
			if(testEntry!=null)
				if(testEntry.getIdentity()!=null) pk = (PageKey)testEntry.getIdentity();
		
			long containerId = pk==null ? -1 : pk.getContainerId().getContainerId();
			long pageNumber = pk==null ? -1 : pk.getPageNumber();
			
			lruDump.write("|("+containerId+","+pageNumber+")");
			
    		}
    	}
    	PageKey pkEntry = (PageKey)entry.getIdentity();
    	lruDump.write(" find::("+pkEntry.getContainerId().getContainerId()+","+pkEntry.getPageNumber()+")");
    	
    	
    	}catch (Exception e) {
			
    		e.printStackTrace();
		}
    	int state = 0x00000000;
    	
    	if(item==null) state+=1;
    	if(item.getEntry()!=entry) state+=2;
    	
    	if(entry!=item.getEntry()) state+=4;
    	
    	if(get(entry.getIdentity())==null) state+=8;
    	if(get(entry.getIdentity())!=item) state+=16;
    	
    	if(entry.getIdentity()==null) state+=32; 
    	
    	if(!entry.getIdentity().equals(key)) state+=64;
    	if(((LRUItem)get(key)).getEntry()!=entry) state+=128;
    	
    	if(item.keepCount<1) state+=256;
    	
    	if(lru.size()>MAX_ENTRIES+20) state+=512;
    	
    	if(state!=0){
    		System.out.println("Error state (find):"+state);//Integer.toBinaryString(state));
    		throw StandardException.newException("Error in TestLRU, find");
    	}
    	
    	return entry;
    }
    
    /**
     * @return the current maximum size of the cache.
     */

    public long getMaximumSize()
    {
        return MAX_ENTRIES;
    }
    /**
     * Change the maximum size of the cache. 
     * @param newSize the new maximum cache size
     */

    public void resize(long newSize)
        throws StandardException
    {
    	System.out.println("resize");
    	
        long currentSize = lru.size();
        if(newSize == currentSize)
            return;
        if(newSize > currentSize)
        {
            for(int i = 0; (long)i < newSize - currentSize; i++){
                LRUItem item = new LRUItem();
            	item.setEntry(holderFactory.newCacheable(this));
            	lru.addLast(item);
            }
        } else
        if(newSize < currentSize)
        {
            Cacheable entry = null;
            do
            {
            	synchronized (this) {

	            	for(int i=0;i<lru.size();i++){
	                LRUItem item = (LRUItem)lru.removeLast();
	                entry = item.getEntry();
	                if(entry == null)
	                    continue;
	                if(entry.isDirty())
	                    entry.clean(false);
	                entry.clearIdentity();
	                item =null;
	            	}
            	}try{
            		wait();
            	}catch(InterruptedException ire){
            		throw StandardException.interrupt(ire);
            	}
            }while(lru.size()>newSize);
            notifyAll();

        }
    }

   /**
    * 
    * Find an object already in cache. If not in cache 
    * return null. 
    * NOTE - Have never seen this method called.
    * @param key - Pagekey of Cacheable we seek.
    * @return Cacheable found.
    */

    public Cacheable findCached(Object key)
        throws StandardException
    {
    	if (debug)System.out.println("findCached");
        LRUItem item = (LRUItem)get(key);
        if(!active)
            return null;
        if(item != null)
        {
            item.incKeep();
            lru.remove(item);
            lru.addFirst(item);
            return item.getEntry();
        } else
        {
            return null;
        }
    }
    
    /**
     * Determine whether a key is in the cache.
     *
     *@param key - Pagekey of Cacheable we seek
     *@return true if found.
     */

    public boolean containsKey(Object key)
    {
    	System.out.println("containskey");
        LRUItem item = (LRUItem)get(key);
        return item != null;
    }
    
    /**
     * 
     * Not implemented, never seen in use
     */
    
    public void setUsed(Object aobj[])
    {
    	System.out.println("setUsed");
    }
    
    /**
	* Release a Cacheable object previously found with find(). 
	* Releasing a page will decrease its keepCount.
	*@param entry - Page to be released.
*/

    public void release(Cacheable entry)
    {
    	if(debug)System.out.println("release");
        LRUItem item = (LRUItem)get(entry.getIdentity());
        int keepCount = item.getKeepCount();
        
        ///
        synchronized(this)
        {
        	item.decKeep();
        	if(SanityManager.DEBUG){
        		
        		if(debug)verifyRelease(entry,item,keepCount);
        	}
        }
    }

    /**
     * Verification method for release.
     * 
     * @param entry - Cacheable to be released
     * @param item - LRUItem holding entry
     * @param keep - keepCount before release
     */
    
    public void verifyRelease(Cacheable entry, LRUItem item,int keep){
    	boolean ok = true;
    	
  	if(keep!=(item.getKeepCount()+1)) ok=false;
    	
    	if(item.getEntry()!=null){
    		if(entry != item.getEntry()) ok =false;
    		if(get(entry.getIdentity())!=null) ok =false;
    		
    	}
    	
    	if(!ok){
    		StandardException.newException("Error in TestLRU, release");
    	}
    }
    
    /**
     * 
     * Explicitly remove a page from cache. Sets 
     * RemoveRequested so no one else will try to remove 
     * the same item.
     * 
     * @param entry - page to be removed
     * 
     * 	@exception StandardException Standard Cloudscape 
     * error policy.
*/
    public void remove(Cacheable entry)
        throws StandardException
    {
    	if(debug)System.out.println("remove");
    	
        LRUItem item = (LRUItem)get(entry.getIdentity());
        item.decKeep();
        if(item.isRemoveRequested())
            return;
        synchronized(this)
        {
       
            item.setRemoveRequested(true);
            do
            {
                if(item.isRemoveRequested())
                    break;
                try
                {
                    wait();
                }
                catch(InterruptedException ire)
                {
                    throw StandardException.interrupt(ire);
                }
            } while(item.getKeepCount() > 0);
            
            if(entry.isDirty())
                entry.clean(true);
            
            remove(item.getEntry().getIdentity());
            entry.clearIdentity();
           
            item.setRemoveRequested(false);
        
        }
    }
    /**
	* Cleaning all items the cache. Passes the call to 
	* the generic cleanCache method.
	*
	* @see Cacheable#clean
	* @see Cacheable#isDirty
	*
	* @exception StandardException Standard Cloudscape 
	* error policy.
*/

    public void cleanAll()
        throws StandardException
    {
    	System.out.println("cleanAll");
        stat.cleanAll++;
        cleanCache((Matchable)null);
    }
    
    /**
	* Clean all objects that match the partialKey (or exact key).
	* Passes the call to the generic cleanCache method.
	*
	*@see Matchable
	@exception StandardException Standard Cloudscape error policy.
*/



    public void clean(Matchable partialKey)
        throws StandardException
    {
    	if(debug)System.out.println("clean");
        cleanCache(partialKey);
    }
    
    /**
	* Age as many objects as possible out of the cache. 
	* Method not implemented. This should not have an 
	* impact on cache manager performance. As far as we 
	* have observed the method is only called during shutdown
	*
	*@see Cacheable#clean
	*@see Cacheable#clearIdentity
*/

    public void ageOut()
    {
    	
    	if(debug)System.out.println("ageOut");
    	try {
			f = new FileWriter("Test.txt");
			f.write("" + stat);
			f.close();
		} catch(Exception e){
			System.out.println("Error med skriving til fil");
			e.printStackTrace();
		}
    }
    
    /**
	Shutdown the cache. This call stops the cache
	returning any more valid references on a find() or 
	findCached() call, 	and then cleanAll() and ageOut()
	 are called. The cache remains in existence until 
	 the last kept object has been unkept.

	@exception StandardException Standard Cloudscape error policy.

*/


    public void shutdown()
        throws StandardException
    {
    	System.out.println("shutDown");
        synchronized(this)
        {
            active = false;
        }
        cleanAll();
        ageOut();
    }

    /**
	*This cache can use this DaemonService if it needs some work to be done
	*in the background.
	*Not implemented
*/
    public void useDaemonService(DaemonService daemonservice)
    {
    	System.out.println("useDaemonService");
    }
    /**
	*
	* Throw all items that match the partial key out of cache.
	* @see Matchable
	* @return true if discard has successful gotten rid of 
	* all objects that match the partial or exact key.  
	* False if some objects that matches were not gotten 
	* rid of because it was kept.
*/

    public boolean discard(Matchable partialKey)
    {
    	if (debug)System.out.println("discard");
        boolean discardedAll = true;
        synchronized(this)
        {
        	
        	for (Iterator iter = lru.iterator(); iter.hasNext();) {
			
				LRUItem item= (LRUItem) iter.next();
				Cacheable entry = item.getEntry();
				if(entry==null) continue;
				if(item.getKeepCount()==0){
					if(partialKey!=null ){
						if(partialKey.match(entry.getIdentity())){
							remove(entry.getIdentity());
							entry = clearIdentity(entry);
							
						}	
					} else {
						remove(entry.getIdentity());
						entry = clearIdentity(entry);
						
					}
				}
				else discardedAll = false;
			}
        }
        try {
        	FileWriter f = new FileWriter("Test.txt");
			f.write("" + stat);
			f.write("//Cache size: "+MAX_ENTRIES);
			f.close();
		} catch(Exception e){
			System.out.println("Error med skriving til fil");
			e.printStackTrace();
		}
        return discardedAll;
    }
    
    /**
	* Report the number of items in use (with Identity) 
	* in this cache.
 */

    public int getNumberInUse()
    {
    	System.out.println("getNumberInUse");
        int number = 0;
        synchronized(this)
        {
            Iterator iter = lru.iterator();
            do
            {
                if(!iter.hasNext())
                    break;
                LRUItem element = (LRUItem)iter.next();
                if(element.getEntry().getIdentity() != null)
                    number++;
            } while(true);
        }
        return number;
    }
    
    /**
	Return statistics about cache that may be implemented.
**/


    public long[] getCacheStats()
    {
    	System.out.println("getCacheStats");
        return null;
    }


	/**
		reset the cache statistics to 0.
	**/

    public void resetCacheStats()
    {
    	System.out.println("resetCacheStats");
    }
    /**
     * Perform an operation on (approximately) all entries 
     * that matches the filter, or all entries if the filter 
     * is null. Entries that are added while the
     * cache is being scanned might or might not be missed.
     *
     * @param filter
     * @param operator
     */

    public void scan(Matchable filter, Operator operator)
    {
    	System.out.println("scan");
        Cacheable entry = null;
        synchronized(this)
        {
            for(Iterator iter = lru.iterator(); iter.hasNext();)
            {
                LRUItem element = (LRUItem)iter.next();
                entry = element.getEntry();
                Object key = entry.getIdentity();
                if(filter == null || filter.match(key));
            }

        }
        operator.operate(entry);
    }


    //// Methods from Serviceable interface are not implemented. ///////////////////////////
   ////  As far as we can see they are not called. May they are called if the
    ///  database is left on for a longer period of time.

    public int performWork(ContextManager context)
        throws StandardException
    {
    	System.out.println("performWork");
        return 0;
    }

    public boolean serviceASAP()
    {
    	System.out.println("serviceASAP");
        return false;
    }

    public boolean serviceImmediately()
    {
    	System.out.println("serviceImmediatly");
        return false;
    }

    /// End of interface methods //////////////////////////////////////////////////

    public void cleanCache(Matchable partialKey)
        throws StandardException
    {
         synchronized(this)
        {
            Iterator iter = lru.iterator();
            do
            {
                if(!iter.hasNext())
                    break;
                LRUItem element = (LRUItem)iter.next();
                Cacheable entry = element.getEntry();
                if(entry != null&&entry.getIdentity()!=null)
                {
                    Object key = entry.getIdentity();
                    if(partialKey!=null)
                    	if(!partialKey.match(key)) break;
                   	entry.clean(false);
                    
                }
            } while(true);
        }
       
    }
    
    /**
     * Method containing the logic to remove items from the 
     * lru list. The list will have to be traversed to find 
     * an LRUItem with keepCount =0. If no items in the list 
     * have keepCount = null the list will have to be 
     * temporably extended by one (cannot wait for an item 
     * to be released as this would break synchronization). 
     * Later the list will be shortened to its standard 
     * length.
     * 
     * 
     * @param ll - Linked list we want to find a free 
     * 	item in.
     * @param maxSize - The standard size of ll. If real 
     * 	size is more we will try to shorten ll.
     * @param hm - HashMap containing ll's active items.
      * @return LRUItem found
     * @throws StandardException
     */

    public LRUItem findFreeItem(LinkedList ll,int maxSize,HashMap hm) throws StandardException
    {
    	if(debug) System.out.println("findFreeItem");
    	if(debug) System.out.println("LL : "+ll.size());
    	Cacheable entry = null;
    	LRUItem tmp = null;
    	LRUItem element = null;
    	
    	synchronized (this) {
    		
    		/// Traverse the list. Find the last item that can be reclaimed.
        		
            for (int f=ll.size()-1;f>=0;f--) {
            	tmp = (LRUItem) ll.get(f);
				entry = tmp.getEntry();
				if(tmp.isRemoveRequested()) continue;
				if(tmp.getKeepCount() == 0){
					
					if(entry==null||entry.getIdentity()==null){
            			element=tmp;
            			break;
            		}
					
					if(entry.isDirty()) entry.clean(false);
					
					Object o = hm.remove(entry.getIdentity());
					
					
					if(SanityManager.DEBUG){
					
						SanityManager.ASSERT(o!=null,"Error in findFreeItem, no hash-entry");
						if(log){
							try {
								PageKey pk = (PageKey)entry.getIdentity();
								long containerId = pk==null ? -1 : pk.getContainerId().getContainerId();
								long pageNumber = pk==null ? -1 : pk.getPageNumber();
								
								lruDump.write(" evicted:("+containerId+","+pageNumber+")");
							} catch (IOException e) {
								e.printStackTrace();
							}
						}
					}
					element = tmp;
					break;
				}
				element = null;
				tmp = null;
			
            }
        
	        // Did we find an available item?? If not go to list extension.
	        if(element!=null){
	        	
	        	//  If list is longer than maxValue (someone extended 
	        	//  the list because there were no unkept items)
	        	//  we will try to shorten the list.
	        	
	        	// How many items will we try to remove :
	        	
	        	int toLongby = ll.size()-maxSize ;
	        	
	        	tmp = null;
	        	if(toLongby>0){
	        		
	        		for(int i=ll.size()-1;i>=0;i--){
	        			
	        			if(toLongby==0) break;
	        			
	        			tmp = (LRUItem)ll.get(i);
	        			if(tmp==element) continue;
	        			if(tmp.isRemoveRequested()) continue;
	        		
	        			entry=tmp.getEntry();
		        		if(entry!=null){
		        			if(entry.getIdentity()!=null){
		        				if(tmp.getKeepCount()==0){
		        					Object o = hm.remove(entry.getIdentity());
		    						if(SanityManager.DEBUG)	
		    							SanityManager.ASSERT(o!=null,"Error in findFreeItem, no hash-entry");
		    						
		    						if(entry.isDirty())entry.clean(false);
		    						entry = clearIdentity(entry);
		        				}
		        				else continue;
		        			}
		        		}
		        		ll.remove(tmp);
	        			tmp = null;
	        			toLongby--;
	        		}
	        	}
	        	// return the item
	        	return element;
	        }
	        
	        // There are no free items in the list. We have to extend the list by one.
	        if (debug) System.out.println("No free item found, extending the list");
	        element = new LRUItem();
	        ll.add(element);
	    	return element;
        }
    }

    /**
     * Encapsulating the task of creating identiy for a 
     * Cacheable. This method is used if the Cacheable has 
     * no previous identity.
     * 
     * @param entry - the entry to get an identity
     * @param createParameter 
     * @return Cacheable with identity
     */
    
    public Cacheable createIdentity(Cacheable entry,Object key,Object createParameter)throws StandardException{
    	
    	Cacheable retr = entry.createIdentity(key,createParameter);
    	usedItems++;
    	
    	return retr;
    }
    
    /**
     * Encapsulating the task seting identity for a 
     * cacheable. This method is used if the Cacheable has a 
     * previous identity, replacing it with a new one.
     * 
     * @param entry - Entry to get a new identity
     * @param key - Pagekey that will be the new identity of 
     * 	entry
     * @return Cacheable with new identity
     * @throws StandardException
     */
    
    public Cacheable setIdentity(Cacheable entry,Object key) throws StandardException{
		
    	System.out.println("setIdentity");
    	if(entry==null) {
    		entry = holderFactory.newCacheable(this);
    		
    	}
    	usedItems++;
    	
    	Cacheable oldentry;
    	oldentry = entry;
    	if(entry.getIdentity()!=null){
    		if(oldentry.isDirty()) oldentry.clean(false);
        	oldentry = clearIdentity(oldentry);
    		
    	}
    	entry =oldentry.setIdentity(key);
    	return entry;
	}
  
    /**
     * Encapsulating the task of clearing entry of identity
     * 
     * @param entry - Cacheable to have identity cleared.
     * @return entry with identity cleared.
     */

    public Cacheable clearIdentity(Cacheable entry) {
		
    	entry.clearIdentity();
    	usedItems--;
    	return entry;
	}
    
}
