package org.apache.derby.impl.services.cache;

import java.io.FileWriter;
import java.io.IOException;
import java.util.*;

import org.apache.derby.iapi.error.StandardException;
import org.apache.derby.iapi.services.cache.*;
import org.apache.derby.iapi.services.context.ContextManager;
import org.apache.derby.iapi.services.daemon.DaemonService;
import org.apache.derby.iapi.services.daemon.Serviceable;
import org.apache.derby.iapi.services.sanity.SanityManager;
import org.apache.derby.iapi.store.raw.PageKey;
import org.apache.derby.iapi.util.Matchable;
import org.apache.derby.iapi.util.Operator;
import org.apache.derby.impl.store.raw.data.StoredPage;

/**
 * A cache manager based on LRU/2Q. Maintains 3 lists aIn, aM and
 *  aOut. Also has three hashmaps, one for each of the lists. 
 *  aIn is a FIFO screening que. Its main job is to hold pages
 *  while corelated references to the page are made. aOut stores
 *  only pointerts to pages that have been in aIn. If a pointer
 *  found in aOut is referenced the page is put into aM, the 
 *  main buffer.
 *  
 *  Pages in aIn and aOut are put in holder items called 
 *  LRUItems. Most important methods are find, release and 
 *  findfreeItem.
 *  
 *  Some methods have extra verification methods for debugging 
 *  purposes. Derby embedded debugging catches a lot of bugs.
 *  But this is often bundled in a lot of nested exception
 *  handling, and shutdown of the database. Finding the
 *  root (exeception) for these cascading events is not very easy.
 *  The verification methods are designed to let the designer
 *  know what exactly went wrong.
 * 
 *  Not all interface methods are implemented. As far as we can
 *  see many methods never gets called in a page cache manager
 *  implementation, like this.
 *  
 * @see Servicable
 * @see CacheManager
 * @see LRUItem
 * @see Cacheable
 * 
 */

public class LRU2Q
    implements CacheManager, Serviceable
{

    private String name;
    final long MAX_ENTRIES;
    public final CacheStat stat = new CacheStat();
    private LinkedList aM,aIn,aOut;
    private HashMap aMHash,aInHash,aOutHash;
    private int kIn,kOut,kM;
    
    private CacheableFactory holderFactory;
    private boolean active;
    private int usedItems;
    
    

    /// Fields only used for debuggin and statistics.
    private int counter=0;
    private int lineNr=0;
    
    FileWriter f;
    FileWriter lru2;
    FileWriter findCount;
    int nrFind=0;
    boolean debug=true;
    private boolean log =true;

    /**
     * Constructor - Currently ignores initial size and goes 
     * right to maximum size.
     * 
	 *	@param holderFactory the cacheable object class
	 *	@param name the name of the cache
	 *	@param initialSize the initial number of cachable
	 *  object this cache holds.
	 *	@param maximumSize the maximum size of the cache.  

     */

    public LRU2Q(CacheableFactory holderFactory, String name, int initialSize, long maximumSize, boolean useByteCount)
    {
       this.name = name;
        MAX_ENTRIES = maximumSize;
        stat.initialSize = initialSize;
        stat.maxSize = maximumSize;
        this.holderFactory = holderFactory;
        kIn= (int)(maximumSize *0.25);
        kOut= (int)(maximumSize* 0.50);
        kM = (int)(maximumSize *0.75);
        aInHash = new HashMap();
        aOutHash = new HashMap();
        aMHash = new HashMap();
        initialize(maximumSize);
        active = true;
        usedItems =0;
        try{
        	lru2 = new FileWriter("lru2Q.txt");
        	lru2.close();
        	
        }catch (IOException e) {
			e.printStackTrace();	
		}
    }

    /**
     * 
     * Initializes the lists and creates LRUItems.
     * 
     * @param initialsize - size of cache
     */

    private void initialize(long initialsize)
    {
        aM = new LinkedList();
        int aMLength= (int)(initialsize*0.75);
        for(int i = 0; i < aMLength; i++)
        {
            LRUItem item = new LRUItem();
            aM.add(item);
        }
        
        aIn= new LinkedList();
        for(int i = 0; i < kIn; i++)
        {
            LRUItem item = new LRUItem();
            aIn.add(item);
        }
        aOut= new LinkedList();
    }
    
    ////// Interface methods from CacheManager /////////////
    

    /**
     * All items that will be accessed from the cachemanager
     *  must first be created by this method.
     * The page is put in the front of aIn.
     * 
     * @see PageKey
     * 
     * @param key - Pagekey of page to be created
     * @param createParameter
     */

    public  Cacheable create(Object key, Object createParameter)
        throws StandardException
    {
    	if(debug) System.out.println("create");
        
    	LRUItem aMItem = (LRUItem)aMHash.get(key);
    	LRUItem aInItem = (LRUItem)aInHash.get(key);
    	
    	if(aMItem!= null || aInItem != null)	throw StandardException.newException("XBCA0.S", name, key);
		
        Cacheable entry = null;
        Cacheable oldentry=null;
        synchronized (this) {

        	aInItem = findFreeItem(aIn,kIn,aInHash,aOut,aOutHash,kOut);
            
        	aIn.remove(aInItem);
	        aIn.addFirst(aInItem);
	        
	        entry = aInItem.getEntry();
	        if(entry != null)
	        {
	            if(entry.isDirty())
	                entry.clean(false);
	            if(entry.getIdentity()!=null)
	            	entry.clearIdentity();
	       } 
	        entry = holderFactory.newCacheable(this);
	        
	        oldentry = entry;
	        entry = createIdentity(oldentry,key,createParameter);
	        aInItem.incKeep();
	        
	        aInItem.setEntry(entry);
	       aInHash.put(key, aInItem);
        }
     
        if(SanityManager.DEBUG)
        	return verifyCreate(entry,key,aInItem);
        else return entry;
    	
    }
    
    /**
     * Verification method for create, used for debugging 
     * @param entry - from create
     * @param key - from create
     * @param item - from cre<ate
     * @return - Cacheable to be returned from 
     * 		create (if no errors are found}.
     * @throws StandardException
     */

    public Cacheable verifyCreate(Cacheable entry, Object key, LRUItem item) throws StandardException{
    	
    	int state = 0x000000000000000;
    	
    	if(entry==null) state=+ 1;
    	if(item.getEntry()!=entry) state=+ 2;
    	if(!entry.getIdentity().equals(key)) state=+ 4;
    	if(item.keepCount!=1) state=+ 8;
    	if(entry.getIdentity() == null) state +=16;
    	if(((LRUItem)aInHash.get(key)).getEntry()!=entry) state+=32;
    	if(state!=0){
    		System.out.println("Create Errorstate: "+Integer.toBinaryString(state));
    		throw StandardException.newException("Error in create");
    	}
    	
    	return entry;
    	
    }
    

    /**
     * Find an object in cache or fault in from disk. Method
     * contains, together with findfreeItem, the LRU/2Q 
     * logic. A call to find result in one of four cases.
     * 
     * Case 1 - The object is found in aIn. The page is 
     * returned (and kept), but the state of the cache 
     * manager remains the same. 
     * 
     * Case 2 - The Pagekey is found in aOut. The page 
     * is faulted into the front of aM, and the Pagekey 
     * removed from aOut.
     * 
     * Case 3 - The page is found in aM. Page is moved to 
     * front of aM.
     * 
     * Case 4 - The page is not found in cache. Page is 
     * faulted into the front of aIn. Last item of aIn is 
     * removed, and its Pagekey is put in front on aOut.
     * Last Pagekey in aOut is removed.
     * 
     * 
     *@param key - the key to the object
	 *@return a cacheable object that is kept in the cache.
	 *@exception StandardException Cloudscape Standard 
	 *	error policy
	*/

    public Cacheable find(Object key)
        throws StandardException
    {
    	if(debug) System.out.println("find");
    	nrFind++;
    	Cacheable entry;
    	LRUItem item;
    	// If Derby is in shutdown return null
    	if(!active)
            return null;
    	
    	//Statistikk
    	synchronized (this) {
    		counter++;
    		
    		
    		// Debug code, writing statistics
    		if(debug)
				if(counter == 10) {
					try {
						f = new FileWriter("lru2QStat.txt");
						f.write("" + stat+" \n");
						f.close();
						findCount= new FileWriter("findCount.txt");
						findCount.write("#find "+nrFind);
						findCount.close();
					} catch(Exception e){
						System.out.println("Error med skriving til fil");
						e.printStackTrace();
					}
					counter =0;
				}
		   
    		
    		/// CASE 3 Find page in AM.
			/// Move page to front of AM.
	       
				item = (LRUItem)aMHash.get(key);
				if(item != null) {
					//finnes i aM
					aM.remove(item);
					aM.addFirst(item);
					entry = item.getEntry();
					item.incKeep();
					stat.findHit++;
					return verifyFind(item.getEntry(),item,key,aM,aMHash);
				}
				
				/// CASE 2 Page found in aOut
				/// Page faulted into Am
				
				
				PageKey pk = (PageKey)aOutHash.get(key);
				if(pk != null) {
					item = findFreeItem(aM,kM,aMHash,null,null,0);
					entry = item.getEntry();
					if(entry==null)
						entry = holderFactory.newCacheable(this);
					else if(entry.getIdentity()!=null)entry=clearIdentity(entry);
					
					entry = setIdentity(entry,key);
					item.setEntry(entry);
					item.incKeep();
					
					aM.remove(item);
					aM.addFirst(item);
					aOutHash.remove(key);
					aMHash.put(key,item);
					aOut.remove(key);
					stat.findMiss++;
					return verifyFind(entry,item,key,aM,aMHash);
				}
				// Case 1 - Item found in aIn, do nothing
				item = (LRUItem)aInHash.get(key);
				if(item != null) {
					item.incKeep();
					stat.findHit++;
					return item.getEntry();
				}
				
				// CASE 4 - Item not found in cache.
				// Fault in and put in front of Ain.
				item = findFreeItem(aIn,kIn,aInHash,null,null,0);
				
				entry = item.getEntry();
				PageKey toOut;
				if(entry!=null){
					if(entry.getIdentity()!=null){
						toOut = (PageKey)entry.getIdentity();
						aOut.addFirst(toOut);
						
						if(aOut.size()>kOut)
							aOutHash.remove((PageKey)aOut.removeLast());
						aOutHash.put(toOut,toOut);

					}
					entry =	clearIdentity(entry);
				} else {
					entry = holderFactory.newCacheable(this);
				}

				entry =setIdentity(entry,key);
				
				aInHash.put(key,item);
				aIn.remove(item);
				aIn.addFirst(item);
				
				stat.findMiss++;
				item.incKeep();
				item.setEntry(entry);
							
				if(debug) return verifyFind(entry,item,key,aIn,aInHash);
				return entry;
	    	}
    }

    /**
     * 
     * Verification method for find.
     * 
     * @param entry - Cacheable to be returned 
     * @param item - LRUItem holding entry
     * @param key - Pagekey of entry
     * @param ll - LinkedList where page was put
     * @param hm -  HashMap for ll
     * @return returning entry
     * @throws StandardException
     */

    public Cacheable verifyFind(Cacheable entry,LRUItem item,Object key,LinkedList ll,HashMap hm) throws StandardException{

    	if(debug) System.out.println("used items: "+usedItems);
    
    	if(log){
	    	try{
	    		synchronized (this){
	    			
	    			PageKey pk1 = (PageKey) entry.getIdentity();
	        		if(pk1.getContainerId().getContainerId()==896)
	        			if(pk1.getPageNumber()==3) entry.toString();
	        		
	    		
	    			lru2 = new FileWriter("lru2Q.txt",true);
	    			lru2.write("\n");
		    		Cacheable testEntry;
		    		lru2.write("line:"+lineNr+" ");
		    		lineNr++;
		    	
		    		lru2.write(" aIn: ");
		    		for (Iterator iter = aIn.iterator(); iter.hasNext();) {
		    			LRUItem lruItem = (LRUItem) iter.next();
		    			testEntry=lruItem.getEntry();
		    			
		    			PageKey pk=null;
		    			if(testEntry!=null)
		    				if(testEntry.getIdentity()!=null) pk = (PageKey)testEntry.getIdentity();
		    		
		    			long containerId = pk==null ? -1 : pk.getContainerId().getContainerId();
		    			long pageNumber = pk==null ? -1 : pk.getPageNumber();
		    			
		    			lru2.write("|("+containerId+","+pageNumber+")");
		    			
		        		}
		
		    		lru2.write("|| aOUT: ");
		    		for (Iterator iter = aOut.iterator(); iter.hasNext();) {
		    			PageKey pk = (PageKey) iter.next();
		    		
		    			long containerId = pk==null ? -1 : pk.getContainerId().getContainerId();
		    			long pageNumber = pk==null ? -1 : pk.getPageNumber();
		    			
		    			lru2.write("|("+containerId+","+pageNumber+")");
		    			
		        		
		    		}
		    		lru2.write("|| aM: ");
				
		    		for (Iterator iter = aM.iterator(); iter.hasNext();) {
					LRUItem lruItem = (LRUItem) iter.next();
					testEntry=lruItem.getEntry();
					
					PageKey pk=null;
					if(testEntry!=null)
						if(testEntry.getIdentity()!=null) pk = (PageKey)testEntry.getIdentity();
				
					long containerId = pk==null ? -1 : pk.getContainerId().getContainerId();
					long pageNumber = pk==null ? -1 : pk.getPageNumber();
					
					lru2.write("|("+containerId+","+pageNumber+")");
					
		    		}
		    		
			    	PageKey pkEntry = (PageKey)entry.getIdentity();
			    	lru2.write(" |find::("+pkEntry.getContainerId().getContainerId()+","+pkEntry.getPageNumber()+")");
		    	}
	    		lru2.close();
	    	
	    	
	    	}catch (Exception e) {
				
	    		e.printStackTrace();
			}
    	}
		
    	// Check if number of active items in aIn is the 
    	// same as in aInHash
    	
    	int activeItemsInAin=0;
    	for(int i =0;i<aIn.size();i++){
    		
    		LRUItem tmpItem = (LRUItem)aIn.get(i);
    		Cacheable tmpEntry = tmpItem.getEntry();
    		if(tmpEntry!=null)
    			if(tmpEntry.getIdentity()!=null)
    				activeItemsInAin++;
    	}
    	// Check if number of active items in aM is the 
    	// same as in aMHash
    	
    	int activeItemsInAm=0;
    	for(int i =0;i<aM.size();i++){
    		
    		LRUItem tmpItem = (LRUItem)aM.get(i);
    		Cacheable tmpEntry = tmpItem.getEntry();
    		if(tmpEntry!=null)
    			if(tmpEntry.getIdentity()!=null)
    				activeItemsInAm++;
    	}
    	
    	
    	int state = 0x00000000;
    	
    	if(item==null) state+=1;
    	if(item.getEntry()!=entry) state+=2;
    	
    	if(entry!=item.getEntry()) state+=4;
    	
    	if(hm.get(entry.getIdentity())==null) state+=8;
    	if(hm.get(entry.getIdentity())!=item) state+=16;
    	
    	if(entry.getIdentity()==null) state+=32;  

    	if(!entry.getIdentity().equals(key)) state+=64;
    	if(((LRUItem)hm.get(key)).getEntry()!=entry) state+=128;
    	
    	if(item.keepCount<1) state+=256;
    	
    	if(debug){ 
    		System.out.println("aInHash: "+aInHash.keySet().size()+" aIn: "+activeItemsInAin);
	    	System.out.println("aMHash: "+aMHash.keySet().size()+" aM: "+activeItemsInAm);
	    	System.out.println("aOutHash: "+aOutHash.keySet().size()+" aOut: "+aOut.size());
    	}
    	
    	if(state!=0){
    		SanityManager.ASSERT(state==0,"Error state (find):"+state);
    	}
    	return entry;
    }
    /**
     * @return the current maximum size of the cache.
     */

    public long getMaximumSize()
    {
        return MAX_ENTRIES;
    }

    /**
     * Change the maximum size of the cache. 
     * This method is not implemented. As far as we can see 
     * this method does is never called for a page 
     * cache manager.
     * @param newSize the new maximum cache size
     */

    public void resize(long newSize)
        throws StandardException
    {
    	if(debug) System.out.println("resize");
    	
    }

    /**
    * Find an object in the cache. Return null if object not 
    * in cache. This method is not implemented. As far as 
    * we can see this method does is never called for 
    * a page cache manager.
    */

    public Cacheable findCached(Object key)
        throws StandardException
    {
    	if(debug) System.out.println("findCached");
    	return null;
    }
    
    /**
     * Determine whether a key is in the cache.
     * This method is not implemented. As far as we can see 
     * this method does is never called for a page cache 
     * manager.
     */

    public boolean containsKey(Object key)
    {
    	if(debug) System.out.println("containskey");
        LRUItem item = (LRUItem)aInHash.get(key);
        LRUItem item2 = (LRUItem)aMHash.get(key);
        if(item != null)
        	return true;
        else if(item2 != null)
        	return true;
        else return false;
    }
    

    /**
     * Mark a set of entries as having been used. 
     * This method is not implemented. As far as we can see 
     * this method does is never called for a page 
     * cache manager.
     *
     * @param keys the key of the used entry.
     */

    

    public void setUsed(Object aobj[])
    {
    	System.out.println("setUsed");
    }
    
    /**
	* Release a Cacheable object previously found with 
	* find(). Releasing a page will decrease 
	* its keepCount.
	*@param entry - Page to be released.
*/
    public void release(Cacheable entry) 
    {
    	if(debug) System.out.println("release");
    	synchronized(this)
        {
        
    	LRUItem item = (LRUItem)aInHash.get(entry.getIdentity());
        if( item == null)
        	item = (LRUItem)aMHash.get(entry.getIdentity());
        
        int keepCount = item.getKeepCount();
        	item.decKeep();
        	notifyAll();
        	if(SanityManager.DEBUG){
        		SanityManager.ASSERT(!((aInHash.get(entry.getIdentity())!=null)
        				&&(aMHash.get(entry.getIdentity())!=null)),"Error, item in both AIn and AOut");
        		
        		verifyRelease(entry,item,keepCount);
        	}
        }
    }
    
    
    /**
     * Verification method for release.
     * 
     * @param entry - Cacheable to be released
     * @param item - LRUItem holding entry
     * @param keep - keepCount before release
     */

    public void verifyRelease(Cacheable entry, LRUItem item,int keep){
    	boolean ok = true;
    	
    	
    	if(keep!=(item.getKeepCount()+1)) ok=false;
    	
    	if(item.getEntry()!=null){
    		if(entry != item.getEntry()) ok =false;
    		
    	}
    	
    	if(!ok){
    		StandardException.newException("Error in TestLRU, release");
    	}
    }


    /**
     * 
     * Explicitly remove a page from cache. Sets 
     * RemoveRequested so no one else will try to 
     * remove the same item.
     * 
     * @param entry - page to be removed
     * 
     * 	@exception StandardException Standard Cloudscape 
     * error policy.
*/


    public void remove(Cacheable entry)
        throws StandardException
    {
    	if(debug) 
    		System.out.println("remove");
    	synchronized(this)
        {
	    	boolean inAIN;
	        LRUItem item = (LRUItem)aInHash.get(entry.getIdentity());
	        if(item == null) {
	        	item = (LRUItem)aMHash.get(entry.getIdentity());
	        	inAIN = false;
	        } else inAIN = true;
	       

	        item.decKeep(); 	
 	        if(item.isRemoveRequested())
	            return;
	        item.setRemoveRequested(true);

            do
            {
               // Will this work ? Or will waiting break the method synchronization
            	System.out.println("remove looping");
            	try
                {
                    wait();
                }
                catch(InterruptedException ire)
                {
                    throw StandardException.interrupt(ire);
                }
                
            } while(item.getKeepCount() > 0);
            
            if(entry.isDirty())
                entry.clean(true);
            if(inAIN)
            	aInHash.remove(item.getEntry().getIdentity());
            else aMHash.remove(item.getEntry().getIdentity());
            entry.clearIdentity();
            item.setRemoveRequested(false);
        
        }
    }
    

    /**
	* Cleaning all items the cache. Passes the call 
	* to the generic cleanCache method.
	*
	@see Cacheable#clean
	@see Cacheable#isDirty

	@exception StandardException Standard Cloudscape error policy.
*/


    public void cleanAll()
        throws StandardException
    {
    	if(debug) System.out.println("cleanAll");
        stat.cleanAll++;
        cleanCache((Matchable)null);
    }
    /**
	* Clean all objects that match the partialKey.
	* Passes the call to the generic cleanCache method.
	*
	* @see Matchable
	* @exception StandardException Standard Cloudscape 
	* error policy.
*/

    public void clean(Matchable partialKey)
        throws StandardException
    {
    	if(debug) System.out.println("clean");
        cleanCache(partialKey);
    }
  

    /**
	* Age as many objects as possible out of the cache. 
	* Method not implemented. This should not have an impact
	* on cache manager performance. As far as we
	* have observed the method is only called during shutdown
	*
	*@see Cacheable#clean
	*@see Cacheable#clearIdentity
*/


    public void ageOut()
    {
    	
    	if(debug) System.out.println("ageOut");
    	
    	if(debug)
    		try {
				f = new FileWriter("Test.txt");
				f.write("" + stat);
				f.close();
			} catch(Exception e){
				System.out.println("Error med skriving til fil");
				e.printStackTrace();
			}

    }
    

    /**
     * 
     * Shutting down the cache. Cache will not return 
     * any more pages.
     * 
	@exception StandardException Standard Cloudscape error policy.

*/


    public void shutdown()
        throws StandardException
    {
    	System.out.println("shutDown");
        
        active = false;
        
        cleanAll();
        ageOut();
    }

    /**
     * 
     * This cache can use this DaemonService if it needs 
     * some work to be done in the background. Method 
     * not implemented.
     * 
	 
*/
    public void useDaemonService(DaemonService daemonservice)
    {
    	System.out.println("useDaemonService");
    }
    /**
	*
	* Throw all items that match the partial key out 
	* of cache.
	* @see Matchable
	* @return true if discard has successful gotten
	* 	 rid of all objects that match the partial 
	* 	 or exact key. False if some objects that matches
	* 	 were not gotten rid of because it was kept.
*/


    public boolean discard(Matchable partialKey)
    {
    	if(debug) System.out.println("discard");
        boolean discardedAll = true;
        synchronized(this)
        {
        	
        	for (Iterator iter = aIn.iterator(); iter.hasNext();) {
			
				LRUItem item= (LRUItem) iter.next();
				Cacheable entry = item.getEntry();
				if(entry==null) continue;
				if(item.getKeepCount()==0){
					if(partialKey!=null ){
						if(partialKey.match(entry.getIdentity())){
							aInHash.remove(entry.getIdentity());
							entry = clearIdentity(entry);
						}	
					} else {
						aInHash.remove(entry.getIdentity());
						entry = clearIdentity(entry);
					}
				}
				else discardedAll = false;
			}
        	
        	for (Iterator iter = aM.iterator(); iter.hasNext();) {
    			
				LRUItem item= (LRUItem) iter.next();
				Cacheable entry = item.getEntry();
				if(entry==null) continue;
				if(item.getKeepCount()==0){
					if(partialKey!=null ){
						if(partialKey.match(entry.getIdentity())){
							aMHash.remove(entry.getIdentity());
							entry = clearIdentity(entry);
						}	
					} else {
						aMHash.remove(entry.getIdentity());
						entry = clearIdentity(entry);
					}
				}
				else discardedAll = false;
			}
        }
    
        return discardedAll;
    }
    
    public boolean verifyDiscard(boolean discardedAll){
    	
    	boolean ok=true;
    		
    	return discardedAll;
    	
    }
    
    /**
	* Report the number of items in use (with Identity) in 
	* this cache.
 */

    public int getNumberInUse()
    {
    	if(debug) System.out.println("getNumberInUse");
        int number = 0;
        synchronized(this)
        {
            Iterator iter = aIn.iterator();
            do
            {
                if(!iter.hasNext())
                    break;
                LRUItem element = (LRUItem)iter.next();
                if(element.getEntry().getIdentity() != null)
                    number++;
            } while(true);
            
            iter = aM.iterator();
            do
            {
                if(!iter.hasNext())
                    break;
                LRUItem element = (LRUItem)iter.next();
                if(element.getEntry().getIdentity() != null)
                    number++;
            } while(true);

        }
        return number;
    }
    
    /**
	* Return statistics about cache that may 
	* be implemented.
**/


    public long[] getCacheStats()
    {
    	System.out.println("getCacheStats");
        return null;
    }


	/**
		reset the cache statistics to 0.
	**/

    public void resetCacheStats()
    {
    	System.out.println("resetCacheStats");
    }
    /**
     * Perform an operation on (approximately) all entries 
     * that matches the filter, or all entries if the filter
     * is null. Entries that are added while the
     * cache is being scanned might or might not be missed.
     *
     * @param filter
     * @param operator
     */

    public void scan(Matchable filter, Operator operator)
    {
    	if(debug) System.out.println("scan");
        Cacheable entry = null;
        synchronized(this)
        {
            for(Iterator iter = aIn.iterator(); iter.hasNext();)
            {
                LRUItem element = (LRUItem)iter.next();
                entry = element.getEntry();
                Object key = entry.getIdentity();
                if(filter == null || filter.match(key))
                	operator.operate(entry);
            }

            for(Iterator iter = aM.iterator(); iter.hasNext();)
            {
                LRUItem element = (LRUItem)iter.next();
                entry = element.getEntry();
                Object key = entry.getIdentity();
                if(filter == null || filter.match(key))
                	operator.operate(entry);
            }

        }

    }

/*
   Methods from Serviceable interface are not implemented. 
   As far as we can see they are not called. May they are 
   called if the database is left on for a longer 
   period of time.
*/
    public int performWork(ContextManager context)
        throws StandardException
    {
    	System.out.println("performWork");
        return 0;
    }

    public boolean serviceASAP()
    {
    	System.out.println("serviceASAP");
        return false;
    }

    public boolean serviceImmediately()
    {
    	System.out.println("serviceImmediatly");
        return false;
    }
 
    /// End of interface methods //////////////////////////

    /**
     * 
     * Method clean the cache for items matching a partial key. 
     * If the partial key is null, all items are cleaned.
     * 
     * 
     * @see Matchable
     */
    


    public void cleanCache(Matchable partialKey)
        throws StandardException
    {
        synchronized(this)
        {
            Iterator iter = aIn.iterator();
            do
            {
                if(!iter.hasNext())
                    break;
                LRUItem element = (LRUItem)iter.next();
                Cacheable entry = element.getEntry();
                if(entry != null&&entry.getIdentity()!=null)
                {
                    Object key = entry.getIdentity();
                    if(partialKey!=null)
                    	if(!partialKey.match(key)) continue;
                   	entry.clean(false);
                    
                }
            } while(true);
            
            iter = aM.iterator();
            do
            {
                if(!iter.hasNext())
                    break;
                LRUItem element = (LRUItem)iter.next();
                Cacheable entry = element.getEntry();
                if(entry != null&&entry.getIdentity()!=null)
                {
                    Object key = entry.getIdentity();
                    if(partialKey!=null)
                    	if(!partialKey.match(key)) continue;
                   	entry.clean(false);
                    
                }
            } while(true);
        }
      
    }

    /**
     * Method containing the logic to remove items from page
     * cache lists. Lists will have to be traversed to find 
     * an LRUItem with keepCount =0. If no items in the list 
     * have keepCount = null the list will have to be 
     * temporably extended by one (cannot wait for an item 
     * to be released as this would break synchronization). 
     * Later the list will be shortened to its standard length.
     * 
     * 
     * @param ll - Linked list we want to find a free 
     * 	item in.
     * @param maxSize - The standard size of ll. If 
     * 	real size is more we will try to shorten ll.
     * @param hm - HashMap containing ll's active items.
     * @param overflow - If called from find, Case 1, 
     * 	this will be aOut. Else null
     * @param OverflowHash - If called from find, Case 1,
     * 	 this will be aOutHash. Else null
     * @param overFlowSize - If called from find, Case 1,
     * 	 this will be kOut. Else null
     * @return a free LRUItem
     * @throws StandardException
     */
    
    public LRUItem findFreeItem(LinkedList ll,int maxSize,HashMap hm, LinkedList overflow,HashMap OverflowHash,int overFlowSize) throws StandardException
    {
    	if(debug) System.out.println("findFreeItem");
    	if(debug) System.out.println("LL : "+ll.size());
    	Cacheable entry = null;
    	LRUItem tmp = null;
    	LRUItem element = null;
    	
    	synchronized (this) {
    		
    	
    		
    		/// Traverse the list. Find the last item that can be reclaimed.
        		
            for (int f=ll.size()-1;f>=0;f--) {
            	tmp = (LRUItem) ll.get(f);
				entry = tmp.getEntry();
				if(tmp.isRemoveRequested()) continue;
				if(tmp.getKeepCount() == 0){
					
					if(entry==null||entry.getIdentity()==null){
            			element=tmp;
            			break;
            		}
					
					if(entry.isDirty()) entry.clean(false);
					
					Object o = hm.remove(entry.getIdentity());
					
					if(SanityManager.DEBUG){
					
						SanityManager.ASSERT(o!=null,"Error in findFreeItem, no hash-entry");
						if(log){
							try {
								lru2 = new FileWriter("lru2Q.txt",true);
								PageKey pk = (PageKey)entry.getIdentity();
								long containerId = pk==null ? -1 : pk.getContainerId().getContainerId();
								long pageNumber = pk==null ? -1 : pk.getPageNumber();
								
								lru2.write(" evicted:("+containerId+","+pageNumber+")");
								lru2.close();
							} catch (IOException e) {
								e.printStackTrace();
							}
						}
					}
					element = tmp;
					break;
				}
				element = null;
				tmp = null;
			
            }
        
	        // Did we find an available item?? If not go to list extension.
	        if(element!=null){
	        	
	        	//  If list is longer than maxValue (someone extended 
	        	//  the list because there were no unkept items)
	        	//  we will try to shorten the list.
	        	
	        	// How many items will we try to remove :
	        	
	        	int toLongby = ll.size()-maxSize ;
	        	
	        	tmp = null;
	        	if(toLongby>0){
	        		
	        		for(int i=ll.size()-1;i>=0;i--){
	        			
	        			if(toLongby==0) break;
	        			
	        			tmp = (LRUItem)ll.get(i);
	        			if(tmp==element) continue;
	        			if(tmp.isRemoveRequested()) continue;
	        		
	        			entry=tmp.getEntry();
		        		if(entry!=null){
		        			if(entry.getIdentity()!=null){
		        				if(tmp.getKeepCount()==0){
		        					Object o = hm.remove(entry.getIdentity());
		    						
		        					if(SanityManager.DEBUG)	
		    							SanityManager.ASSERT(o!=null,"Error in findFreeItem, no hash-entry");
		    						
		    						if(entry.isDirty())entry.clean(false);
		    						entry = clearIdentity(entry);
		        				}
		        				else continue;
		        			}
		        			
		        			// If the list we are working with is aIn, we must 
		        			// put the removed element in aOut.
		        			// If the list is aM, just forget the last item
		        			if(overflow!=null){
		        				overflow.addFirst(entry.getIdentity());
		        				OverflowHash.put(entry.getIdentity(),entry.getIdentity());
		        				if(overflow.size()>overFlowSize){
		        					Cacheable rmv = (Cacheable)overflow.removeLast();
		        					if(rmv.getIdentity()!=null)
		        						OverflowHash.remove(rmv.getIdentity());
		        				}
		        				
		        			}
		        			
		        		}
		        	
		        		ll.remove(tmp);
	        			
		        		
		        		tmp = null;
	        			toLongby--;
	        		}
	        	}
	        	// return the item
	        	return element;
	        }
	        
	        // There are no free items in the list. We have to extend the list by one.
	        if (debug) System.out.println("No free item found, extending the list");
	        element = new LRUItem();
	        ll.add(element);
	    	return element;
        }
    }    
    
    /**
     * Encapsulating the task of creating identiy for a 
     * Cacheable. This method is used if the Cacheable has no
     * previous identity.
     * 
     * @param entry - the entry to get an identity
     * @param createParameter 
     * @return Cacheable with identity
     */

   public Cacheable createIdentity(Cacheable entry,Object key,Object createParameter)throws StandardException{
    	
    	Cacheable retr = entry.createIdentity(key,createParameter);
    	usedItems++; 
    	
    	return retr;
    }
    
   /**
    * Encapsulating the task seting identity for a cacheable. 
    * This method is used if the Cacheable has a previous 
    * identity, replacing it with a new one.
    * 
    * @param entry - Entry to get a new identity
    * @param key - Pagekey that will be the new identity 
    * 	of entry
    * @return Cacheable with new identity
    * @throws StandardException
    */
   
    public Cacheable setIdentity(Cacheable entry,Object key) throws StandardException{
		
    	if(debug) System.out.println("setIdentity");
    	if(entry==null) {
    		entry = holderFactory.newCacheable(this);
    		
    	}
    	usedItems++;
    	
    	Cacheable oldentry;
    	oldentry = entry;
    	if(entry.getIdentity()!=null){
    		if(oldentry.isDirty()) oldentry.clean(false);
        	oldentry = clearIdentity(oldentry);
    		
    	}
    	entry =oldentry.setIdentity(key);
    	return entry;
	}
    
    /**
     * Encapsulating the task of clearing entry of identity
     * 
     * @param entry - Cacheable to have identity cleared.
     * @return entry with identity cleared.
     */

    public Cacheable clearIdentity(Cacheable entry) {
		
    	entry.clearIdentity();
    	usedItems--;
    	return entry;
	}
    
}
