package org.apache.derby.impl.services.cache;

import java.io.FileWriter;

import org.apache.derby.iapi.error.StandardException;
import org.apache.derby.iapi.services.cache.Cacheable;
import org.apache.derby.iapi.services.sanity.SanityManager;

/**
 * Holder class for Cacheables, used by the LRU ad LRU/2Q cache managers.
 * The class can hold information the Cacheable can not hold itself.
 * 
 * KeepCount:
 * This is the number of references that are held from outside of the
 * cachemanager to a Cacheable. KeepCount is increased by one when
 * someone accesses the Cacheable through find or create. It is decreased
 * by one when the Cacheable is released.
 * 
 * RemoveRequested:
 * Flagg set when someone tries to remove a Cacheable from the cache. If
 * keepCount is above 0, removing will have to wait. The flagg stops
 * someone else from trying to remove the same Cacheable.
 */

public class LRUItem {
	
	// the keepCount
	int keepCount=0;
	// the item we are holding
	Cacheable entry=null;
	// Have someone tried to remove this Cacheable from the cache
	boolean removeRequested=false;
	/**
	 * Constructor. Do nothing
	 *
	 */	
	
	public LRUItem(){	
	}
	
	/**
	 * Get the keepCount 
	 * 
	 * 
	 * @return the number of references held to the entry.
	 */
	
	public int getKeepCount(){
		
		if(SanityManager.DEBUG)	
			SanityManager.ASSERT(keepCount>=0,"keepCount under 0");
			
			
		return keepCount;
		
	}
	
	/**
	 * Increase keepCount by one
	 */
	public void incKeep(){
		keepCount++;
	}
	/**
	 * Decrease keepCount by one
	 *
	 */
	public void decKeep(){
		keepCount--;
	}

	/**
	 * Get this Cacheable entry
	 * 
	 * @return
	 */
	public Cacheable getEntry() {
		return entry;
	}

	/**
	 * Set a new entry
	 * @param entry - new Cacheable
	 */
	public void setEntry(Cacheable entry) {
		this.entry = entry;
	}

	/**
	 * Check if someone tried to remove the Cacheable
	 * @return
	 */
	public boolean isRemoveRequested() {
		return removeRequested;
	}

	/**
	 * Set the removeRequested
	 * @param remove. true If trying to remove, false when finnished removing
	 */

	public void setRemoveRequested(boolean remove) {
		this.removeRequested = remove;
	}
	
}
